/* AbiWord
 * Copyright (C) 2001 AbiSource, Inc.
 * Copyright (C) 2002-2003 Marc Maurer (j.m.maurer@student.utwente.nl)
 * Copyright (C) 2002-2003 William Lachance (wlach@interlog.com)
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA.
 */

/* See bug 1764
 * "This product is not manufactured, approved, or supported by 
 * Corel Corporation or Corel Corporation Limited."
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <locale.h>
#include <gsf/gsf-utils.h>
#include <gsf/gsf-input-stdio.h>

#include "ut_types.h"
#include "ut_string.h"
#include "ut_string_class.h"
#include "ut_units.h"
#include "ut_growbuf.h"
#include "ut_assert.h"
#include "ut_debugmsg.h"
#include "ut_math.h" // for rint (font size)
#include "ut_rand.h"

#include "xap_Frame.h"
#include "xap_EncodingManager.h"

#include "pd_Document.h"
#include "pt_Types.h"

#include "fl_AutoLists.h"
#include "fl_AutoNum.h"

#include "ap_Strings.h"

#include "ie_imp_WordPerfect.h"
#include "ie_impexp_WordPerfect.h"

#define X_CheckDocumentError(v) if (!v) { UT_DEBUGMSG(("X_CheckDocumentError: %d\n", __LINE__)); }

IE_Imp_WordPerfect_Sniffer::IE_Imp_WordPerfect_Sniffer()
    : IE_ImpSniffer(IE_MIME_WP_6)
{
}

IE_Imp_WordPerfect_Sniffer::~IE_Imp_WordPerfect_Sniffer()
{
}

UT_Confidence_t IE_Imp_WordPerfect_Sniffer::recognizeContents (const char * szBuf, 
							       UT_uint32 iNumbytes)
{
    char * magic    = 0;
    int magicoffset = 0;
   
    magic = "WPC";
    magicoffset = 1;
    
    if (iNumbytes > (magicoffset + strlen (magic)))
    {
	if (!strncmp (szBuf + magicoffset, magic, strlen (magic)))
	{
	    int productType = (int) *(szBuf + WP_HEADER_PRODUCT_TYPE_OFFSET);
	    int fileType = (int) *(szBuf + WP_HEADER_FILE_TYPE_OFFSET);
	    int majorVersion = (int) *(szBuf + WP_HEADER_MAJOR_VERSION_OFFSET);
	    int minorVersion = (int) *(szBuf + WP_HEADER_MINOR_VERSION_OFFSET);
	    UT_DEBUGMSG(("product type: %i, file type: %i, major version: %i, minor version: %i\n", productType, fileType, majorVersion, minorVersion ));
	    // we only want to try parsing wordperfect 6/7/8/9 documents for now
	    if ((majorVersion != WP_WORDPERFECT6789_EXPECTED_MAJOR_VERSION) || (fileType != WP_WORDPERFECT_DOCUMENT_FILE_TYPE))
		return UT_CONFIDENCE_POOR;
	    
	    return UT_CONFIDENCE_PERFECT;
	}
    }
   
	// ok, that didn't work, we'll try to dig through the OLE stream
	char olemagic [36] = // UTF16 encoded "PerfectOffice_Main"
		{0x50, 0x00, 0x65, 0x00, 0x72, 0x00, 0x66, 0x00, 0x65, 0x00, 0x63, 0x00, 0x74, 0x00, 0x4F, 0x00,
		 0x66, 0x00, 0x66, 0x00, 0x69, 0x00, 0x63, 0x00, 0x65, 0x00, 0x5F, 0x00, 0x4D, 0x00, 0x41, 0x00,
		 0x49, 0x00, 0x4E, 0x00};

	// WARNING: very dumb code ahead, but I don't want to think today
	int k = 0;
	for (int i=0; i<iNumbytes && k<36 ; i++)
	{
		if (szBuf[i] == olemagic[k]) 
			{ k++; }
		else 
			{ k = 0; }
			
	}
	if (k == 36)
	{
		// right, that string existed. confidence is Good now... not perfect I guess, since 
		// we might be dealing with a Quatro Pro document using the "PerfectOffice_Main" string
		return UT_CONFIDENCE_GOOD;
	}

    return UT_CONFIDENCE_ZILCH;
}

UT_Confidence_t IE_Imp_WordPerfect_Sniffer::recognizeSuffix (const char * szSuffix)
{
    // We recognize both word documents and their template versions
    if (!UT_stricmp(szSuffix,".wpd") || !UT_stricmp(szSuffix, ".wp"))
	return UT_CONFIDENCE_PERFECT;
    return UT_CONFIDENCE_ZILCH;
}

UT_Error IE_Imp_WordPerfect_Sniffer::constructImporter (PD_Document * pDocument,
							IE_Imp ** ppie)
{
    IE_Imp_WordPerfect * p = new IE_Imp_WordPerfect(pDocument);
    *ppie = p;
    return UT_OK;
}

bool IE_Imp_WordPerfect_Sniffer::getDlgLabels  (const char ** pszDesc,
						const char ** pszSuffixList,
						IEFileType * ft)
{
    *pszDesc = "WordPerfect 6/7/8/9 (.wpd)";
    *pszSuffixList = "*.wpd";
    *ft = getFileType();
    return true;
}

/****************************************************************************/
/****************************************************************************/

IE_Imp_WordPerfect::IE_Imp_WordPerfect(PD_Document * pDocument)
    : IE_Imp (pDocument),
      m_bParagraphChanged(FALSE),
      m_bParagraphInSection(FALSE),
      m_bInSection(FALSE),
      m_bSectionChanged(FALSE)
{
    // section props
    m_bLeftMarginSet = FALSE;
    m_bRightMarginSet = FALSE;
    m_leftMargin = 0.0f;
    m_rightMargin = 0.0f;
    m_numColumns = 1;
    m_bColumnsSet = FALSE;
    
    m_bInCell = FALSE;
}

IE_Imp_WordPerfect::~IE_Imp_WordPerfect() 
{
}

UT_Error IE_Imp_WordPerfect::importFile(const char * szFilename)
{
    GError *err;
    GsfInput * input;
    input = GSF_INPUT(gsf_input_stdio_new (szFilename, &err));
    if (input == NULL) 
    {
	g_return_val_if_fail (err != NULL, 1);
	
	g_warning ("'%s' error: %s", szFilename, err->message);
	g_error_free (err);
	return 1;
    }
    
    WPXLLListener * listener = new WP6HLListener(this);
    WPXParser * parser = new WP6Parser(input, listener);
    parser->parse();
    delete parser;
    delete listener;
    
    return UT_OK;
}

void IE_Imp_WordPerfect::pasteFromBuffer (PD_DocumentRange *, 
					  unsigned char *, unsigned int, const char *)
{
    // nada
}

void IE_Imp_WordPerfect::startDocument()
{
    UT_DEBUGMSG(("AbiWordPerfect: startDocument\n"));
    
    //_appendSection();
}

void IE_Imp_WordPerfect::endDocument()
{
    UT_DEBUGMSG(("AbiWordPerfect: endDocument\n"));
}

void IE_Imp_WordPerfect::openParagraph(const guint8 paragraphJustification, const guint32 textAttributeBits,
				       const gchar *fontName, const gfloat fontSize, 
				       const gboolean isColumnBreak, const gboolean isPageBreak)
{
    UT_DEBUGMSG(("AbiWordPerfect: openParagraph(paragraphJustification: %d, textAttributeBits: %d)\n", paragraphJustification, textAttributeBits));
    
	if (isPageBreak)
	{
		UT_UCS4Char ucs = UCS_FF;
		X_CheckDocumentError(getDoc()->appendSpan(&ucs,1));			
	}
	else if (isColumnBreak)
	{
		UT_UCS4Char ucs = UCS_VTAB;
		X_CheckDocumentError(getDoc()->appendSpan(&ucs,1));
	}
	else
	{
		_appendParagraph(paragraphJustification, textAttributeBits, fontName, fontSize);
	}
}

void IE_Imp_WordPerfect::openSpan(const guint32 textAttributeBits, const gchar *fontName, const gfloat fontSize)
{
   UT_DEBUGMSG(("AbiWordPerfect: openSpan(textAttributeBits: %d)\n", textAttributeBits));
   
   _appendSpan(textAttributeBits, fontName, fontSize);
}

void IE_Imp_WordPerfect::openSection(guint numColumns, gfloat marginLeft, gfloat marginRight)
{
	UT_DEBUGMSG(("AbiWordPerfect: openSection(numColumns: %d, marginLeft: %.2f, marginRight: %.2f)\n", numColumns, marginLeft, marginRight));
   
	_appendSection(numColumns, marginLeft, marginRight);
}

void IE_Imp_WordPerfect::insertText(const guint16 *textArray, const guint len)
{
    UT_DEBUGMSG(("AbiWordPerfect: insertText\n"));
    
    if (len)
    {
	UT_UCS4String ucs4string;
	for (guint i=0; i<len; i++)
	{
	    ucs4string += (UT_UCS4Char)textArray[i];
	    printf("%c", textArray[i]);
	}
		X_CheckDocumentError(getDoc()->appendSpan(ucs4string.ucs4_str(), len));
		printf("\n");
    }
}

void IE_Imp_WordPerfect::insertLineBreak()
{
    /*UT_DEBUGMSG(("AbiWordPerfect: insertBreak(breakType: %d\n", breakType));
      
    switch (breakType)
    {
    case WPX_PAGE_BREAK:
    {
    UT_UCS4Char ucs = UCS_FF;
    X_CheckDocumentError(getDoc()->appendSpan(&ucs,1));			
    }
    break;
    case  WPX_COLUMN_BREAK:
    {
    UT_UCS4Char ucs = UCS_VTAB;
    X_CheckDocumentError(getDoc()->appendSpan(&ucs,1));
    }
    break;
    default:
    break;
    }*/
}

void IE_Imp_WordPerfect::openTable()
{
    UT_DEBUGMSG(("AbiWordPerfect: openTable\n"));
    
    X_CheckDocumentError(getDoc()->appendStrux(PTX_SectionTable, NULL));
}

void IE_Imp_WordPerfect::openRow()
{
    UT_DEBUGMSG(("AbiWordPerfect: openRow\n"));
    if (m_bInCell)
    {		
		X_CheckDocumentError(getDoc()->appendStrux(PTX_EndCell, NULL));
    }
    
    m_bInCell = FALSE;
}

void IE_Imp_WordPerfect::openCell(const guint32 col, const guint32 row, const guint32 colSpan, const guint32 rowSpan, const RGBSColor * cellFgColor, const RGBSColor * cellBgColor)
{
    UT_DEBUGMSG(("AbiWordPerfect: openCell(col: %d, row: %d, colSpan: %d, rowSpan: %d\n", col, row, colSpan, rowSpan));
    if (m_bInCell)
    {		
		X_CheckDocumentError(getDoc()->appendStrux(PTX_EndCell, NULL));
    }
    
    UT_String propBuffer;
    UT_String_sprintf(propBuffer, "left-attach:%d; right-attach:%d; top-attach:%d; bot-attach:%d", col, col+colSpan, row, row+rowSpan);
	if (cellFgColor)
	{
		UT_String bgCol;
		// we only support bg-style:1 for now
		UT_String_sprintf(bgCol, "; bg-style:1; bgcolor:%02x%02x%02x", cellFgColor->m_r, cellFgColor->m_g, cellFgColor->m_b);
		propBuffer += bgCol;
	}
	
	// TODO: Handle the Cell Background Color here, but I don't think AbiWord supports that
	// ...
	
	UT_DEBUGMSG(("AbiWordPerfect: Inserting a Cell definition: %s\n", propBuffer.c_str()));
	
	const XML_Char* propsArray[3];
	propsArray[0] = "props";
	propsArray[1] = propBuffer.c_str();
	propsArray[2] = NULL;
	
	X_CheckDocumentError(getDoc()->appendStrux(PTX_SectionCell, propsArray));
	m_bInCell = TRUE;
}

void IE_Imp_WordPerfect::closeTable()
{
    if (m_bInCell)
    {
		X_CheckDocumentError(getDoc()->appendStrux(PTX_EndCell, NULL));
    }
    X_CheckDocumentError(getDoc()->appendStrux(PTX_EndTable, NULL));
    m_bInCell = FALSE;
}

// insert the text in the current textbuf to the document, taking its
// style into account
/*
  UT_Error IE_Imp_WordPerfect::_appendCurrentListProperties()
  { 
  UT_DEBUGMSG(("WordPerfect: LISTS: Appending Current List Properties (outline hash: %i)\n", m_currentOutlineHash));
  
  WordPerfectListDefinition *listDefinition = _getListDefinition(m_currentOutlineHash);
  if (listDefinition == NULL)
  return UT_ERROR;
  
  UT_String szListID;
  UT_String szParentID;
  UT_String szLevel;
  UT_String_sprintf(szListID,"%d",listDefinition->getListID(m_currentListLevel));
  if (m_currentListLevel > 0) 
  UT_String_sprintf(szParentID,"%d", listDefinition->getListID((m_currentListLevel-1)));
  else
  UT_String_sprintf(szParentID,"0"); 
  UT_String_sprintf(szLevel,"%d", (m_currentListLevel+1));

  const XML_Char* listAttribs[PT_MAX_ATTRIBUTES*2 + 1];
  UT_uint32 attribsCount=0;
  
  listAttribs[attribsCount++] = PT_LISTID_ATTRIBUTE_NAME;
  listAttribs[attribsCount++] = szListID.c_str();
  listAttribs[attribsCount++] = PT_PARENTID_ATTRIBUTE_NAME;
  listAttribs[attribsCount++] = szParentID.c_str();
   listAttribs[attribsCount++] = PT_LEVEL_ATTRIBUTE_NAME;
   listAttribs[attribsCount++] = szLevel.c_str();
   
   // Now handle the Abi List properties
   UT_String propBuffer;
   UT_String tempBuffer;
   UT_String_sprintf(tempBuffer,"list-style:%d;", listDefinition->getListType(m_currentListLevel));
   propBuffer += tempBuffer;
   // FIXME: writing the list delimiter is kind of tricky and silly (because wordperfect wants to define
   // it within the document, while abi wants to (sensibly probably) define it in the list definition)
   // (we reset it each time but only for numbered lists)
   if (listDefinition->isLevelNumbered(m_currentListLevel)) 
     {	
	UT_DEBUGMSG(("WordPerfect: Appending this list delim: %s\n", m_rightListDelim.c_str()));
	listDefinition->setListRightDelimText(m_currentListLevel, m_rightListDelim.c_str());
	X_CheckWordPerfectError(_updateDocumentListDefinition(listDefinition, m_currentListLevel));
     }   
   if (listDefinition->getListType(m_currentListLevel) == BULLETED_LIST)
     UT_String_sprintf(tempBuffer, "field-font:Symbol; ");
   else
     UT_String_sprintf(tempBuffer, "field-font:NULL; ");
   propBuffer += tempBuffer;
   UT_String_sprintf(tempBuffer, "start-value:%i; ", listDefinition->getListNumber(m_currentListLevel));
   propBuffer += tempBuffer;
   UT_String_sprintf(tempBuffer, "text-indent:%fin; ", m_paragraphStyleTextIndent); 
   propBuffer += tempBuffer;
   UT_String_sprintf(tempBuffer, "margin-left:%fin", m_paragraphStyleMarginLeft);
   propBuffer += tempBuffer;
   listAttribs[attribsCount++] = PT_PROPS_ATTRIBUTE_NAME;
   listAttribs[attribsCount++] = propBuffer.c_str();
   listAttribs[attribsCount++] = NULL;

   if ( (!m_bInSection) || m_bSectionChanged)
     X_CheckWordPerfectError(_appendSection ());

   X_CheckDocumentError(getDoc()->appendStrux(PTX_Block, listAttribs));   

   UT_DEBUGMSG(("WordPerfect: LISTS - Appended a list attributes section (text-indent: %f, margin-left:%f)\n", m_paragraphStyleTextIndent, m_paragraphStyleMarginLeft));

   // hang text off of a list label
   X_CheckWordPerfectError(_appendCurrentTextProperties());
   getDoc()->appendFmtMark();
   m_currentListTag = 0;
   UT_DEBUGMSG(("WordPerfect: LISTS - Appended a list tag def'n (character props)\n"));
   
   // append a list field label
   const XML_Char* fielddef[3];
   fielddef[0] ="type";
   fielddef[1] = "list_label";
   fielddef[2] = NULL;
   X_CheckDocumentError(getDoc()->appendObject(PTO_Field,fielddef));
   UT_DEBUGMSG(("WordPerfect: LISTS - Appended a field def'n\n"));
      
   X_CheckWordPerfectError(_appendCurrentTextProperties());
   
   m_bParagraphExists = true; // in case we start the document with a list
   m_bParagraphChanged = false; // we already changed the paragraph here, so don't go around saying it hasn't been changed when we flush the text
   
   //m_leftListDelim.clear();
   m_rightListDelim.clear();
   
   return UT_OK;
}
*/

UT_Error IE_Imp_WordPerfect::_appendSpan(const guint32 textAttributeBits, const gchar *fontName, const gfloat fontSize)
{
    UT_DEBUGMSG(("AbiWordPerfect: Appending current text properties\n"));
    
    XML_Char* pProps = "props";
    UT_String propBuffer;
    UT_String tempBuffer;
    
    // bold
    propBuffer += "font-weight:";
    propBuffer += (textAttributeBits & WPX_BOLD_BIT ? "bold" : "normal");
    
    // italic
    propBuffer += "; font-style:";
    propBuffer += (textAttributeBits & WPX_ITALICS_BIT ? "italic" : "normal");
    
    // superscript or subscript
    if ((textAttributeBits & WPX_SUPERSCRIPT_BIT) || (textAttributeBits & WPX_SUBSCRIPT_BIT))
    {
		propBuffer += "; text-position:";
		propBuffer += (textAttributeBits & WPX_SUPERSCRIPT_BIT ? "superscript" : "subscript");
    }
    
	// underline & strike-out
    if ((textAttributeBits & WPX_UNDERLINE_BIT) || (textAttributeBits & WPX_STRIKEOUT_BIT))
    {
		propBuffer += "; text-decoration:";
		propBuffer += (textAttributeBits & WPX_UNDERLINE_BIT ? "underline" : "line-through");
    }
    
    // font face
    if (fontName != NULL)
    {
		propBuffer += "; font-family:";
		propBuffer += fontName;
    }
    
    // font face
    UT_String_sprintf(tempBuffer, "; font-size:%.3fpt", fontSize);
    propBuffer += tempBuffer;
   
    UT_DEBUGMSG(("AbiWordPerfect: Appending span format: %s\n", propBuffer.c_str()));
    const XML_Char* propsArray[5];
    
    propsArray[0] = pProps;
    propsArray[1] = propBuffer.c_str();
    propsArray[2] = NULL;
/*	if (m_bPutativeListHasParagraphNumber) 
	{
	propsArray[2] ="type";
	propsArray[3] = "list_label";
	propsArray[4] = NULL;
	}
*/
    X_CheckDocumentError(getDoc()->appendFmt(propsArray));
    return UT_OK;
}

UT_Error IE_Imp_WordPerfect::_appendParagraph(const guint8 paragraphJustification, const guint32 textAttributeBits,
					      const gchar *fontName, const gfloat fontSize)
{
    UT_String propBuffer;
    
    propBuffer += "text-align:";
    switch (paragraphJustification)
    {
		case WPX_PARAGRAPH_JUSTIFICATION_LEFT:
			propBuffer += "left";
			break;
		case WPX_PARAGRAPH_JUSTIFICATION_RIGHT:
			propBuffer += "right";
			break;
		case WPX_PARAGRAPH_JUSTIFICATION_CENTER:
			propBuffer += "center";
			break;
		case WPX_PARAGRAPH_JUSTIFICATION_FULL:
			propBuffer += "justify";
			break;
    }
    
    UT_DEBUGMSG(("AbiWordPerfect: Appending paragraph properties: %s\n", propBuffer.c_str()));
    const XML_Char* propsArray[3];
    propsArray[0] = "props";
    propsArray[1] = propBuffer.c_str();
    propsArray[2] = NULL;
    X_CheckDocumentError(getDoc()->appendStrux(PTX_Block, propsArray));
    
    _appendSpan(textAttributeBits, fontName, fontSize);
    
    return TRUE;
}

UT_Error IE_Imp_WordPerfect::_appendSection(guint numColumns, gfloat marginLeft, gfloat marginRight)
{
    UT_DEBUGMSG(("AbiWordPerfect: Appending section\n"));
    
    UT_String myProps("") ;
    setlocale(LC_NUMERIC, "C");
    myProps += UT_String_sprintf("columns:%d; page-margin-left:%4.4fin; page-margin-right:%4.4fin", numColumns, marginLeft, marginRight);
    setlocale(LC_NUMERIC, NULL);
    
    XML_Char * propsArray[3];
    propsArray[0] = "props";
    propsArray[1] = (XML_Char*)myProps.c_str() ;
    propsArray[2] = NULL ;
    X_CheckDocumentError(getDoc()->appendStrux(PTX_Section, (const XML_Char**)propsArray));
   
    return TRUE;
}

/*
  UT_Error IE_Imp_WordPerfect::_updateDocumentListDefinition(WordPerfectListDefinition *listDefinition, int level)
  {
  // add a list definition to the current document, if necessary
  if (listDefinition->getListID(level) == 0)
  listDefinition->setListID(level, UT_rand());
  
  // finally, set the document's list identification info..
   // FIXME: AbiWord doesn't support different types of lists being appended to their parents.. so we just have to stick with what level 0 provides
   int startingNumber;
   if (listDefinition->getListType(0) == BULLETED_LIST) // REPLACE WITH (WHEN FIXED IN ABI): if (listDefinition->getListType(level) == BULLETED_LIST)
   startingNumber = 0;
   else
   startingNumber = 1;
   fl_AutoNum * pAuto = getDoc()->getListByID(listDefinition->getListID(level));
   // not in document yet, we should create a list for it
   if (pAuto == NULL) 
   {	
   if (level > 0) 
   {	
	     // REPLACE WITH (WHEN FIXED IN ABI): pAuto = new fl_AutoNum(listDefinition->getListID(level), listDefinition->getListID((level-1)), listDefinition->getListType(level), startingNumber, (XML_Char *)"%L", ".", getDoc());
	     pAuto = new fl_AutoNum(listDefinition->getListID(level), listDefinition->getListID((level-1)), listDefinition->getListType(0), startingNumber, (XML_Char *)listDefinition->getListDelim(level).c_str(), ".", getDoc());
	     }   
	     else
	  pAuto = new fl_AutoNum(listDefinition->getListID(level), 0, listDefinition->getListType(level), startingNumber, (XML_Char *)listDefinition->getListDelim(level).c_str(), ".", getDoc());
	  
	  getDoc()->addList(pAuto);
	  }
	  // we should update what we have
	  else 
	  {
	
	  // ADD (WHEN FIXED IN ABI): pAuto->setListType(listDefinition->getListType(level));
	  pAuto->setStartValue(startingNumber);
	  pAuto->setDelim((XML_Char *)listDefinition->getListDelim(level).c_str());
	  }
	  
	  pAuto->fixHierarchy();
	  
	  return TRUE;
   }
*/

